//
// 21-button input method for stationary chorded keyboard with 2 PSP joysticks
// Left hand: 4 finger buttons and 3 mouse buttons
// Right hand: 4 finger buttons and 4 arrow buttons
// Thumbs: BACK, Shift, Control, Alt, Function, SPACE
// v 4.0 24-Dec-2017
// This work is public domain
//
#include <Keyboard.h>
#include <Mouse.h>

#define _CH0 (char)0
#define _CPS (char)0xC1
#define _F01 (char)0xC2
#define _F02 (char)0xC3
#define _F03 (char)0xC4
#define _F04 (char)0xC5
#define _F05 (char)0xC6
#define _F06 (char)0xC7
#define _F07 (char)0xC8
#define _F08 (char)0xC9
#define _F09 (char)0xCA
#define _F10 (char)0xCB
#define _F11 (char)0xCC
#define _F12 (char)0xCD
#define _PRN (char)0xCE
#define _TAB (char)0xB3
#define _RTN (char)0xB0
#define _ESC (char)0xB1
#define _INS (char)0xD1
#define _DEL (char)0xD4
#define _PUP (char)0xD3
#define _PDN (char)0xD6
#define _HOM (char)0xD2
#define _END (char)0xD5
#define _BSP (char)0xB2

#define _UP (char)0xDA
#define _DOWN (char)0xD9
#define _LEFT (char)0xD8
#define _RIGHT (char)0xD7

#define _CTRL (char)0x80
#define _SHIFT (char)0x81
#define _ALT (char)0x82
#define _GUI (char)0x83
#define _CTRL_R (char)0x84
#define _SHIFT_R (char)0x85
#define _ALT_R (char)0x86
#define _RIGHT_R (char)0x87

#define _MOUSE_DECIMATOR 10

// fill here your passwords
#define _PASSWORD_1 ""
#define _PASSWORD_2 ""
#define _PASSWORD_3 ""
#define _PASSWORD_4 ""
#define _PASSWORD_5 ""
#define _PASSWORD_6 ""
#define _PASSWORD_7 ""
#define _PASSWORD_8 ""
#define _PASSWORD_9 ""
#define _PASSWORD_10 ""
#define _PASSWORD_11 ""
#define _PASSWORD_12 ""
#define _PASSWORD_13 ""
#define _PASSWORD_14 ""
#define _PASSWORD_15 ""

// Uncomment for serial debugging
//#define _DEBUG
//#define _DEBUG_MOUSE
//#define _DEBUG_KEYBOARD
//#define _DISABLE_MOUSE

//
// Safety feature, in case programming goes bananas
//
boolean keyboardActive = false;

//
// Pin assignment
//
const int pot_sensorPins[] = {A3, A2, A0, A1};
const int readPins[] = {15,14,16,10};          // reading pins for each block
const int writePins[] = {3, 5, 2, 4, 6, 7};    // pins to set the key block
const int ledPin1 = 8;                         // LED pin for finger buttons
const int ledPin2 = 9;                         // LED pin for modifiers

//
// arrays are used to map values for each of four potentiometers
//
int pot_Low[]= {0, 0, 0, 0};
int pot_Middle[]= {512, 512, 512, 512};  
int pot_High[]= {1023, 1023, 1023, 1023};  
int pot_Calibrated_Low[]= {50, -50, 20, 20};  
int pot_Calibrated_High[]= {-50, 50, -20, -20};  
signed char pot_Position[4];

//
// arrays are used to map values for each 21 buttons and 2 pedals
//
int left[]    = {0, 0, 0, 0};                  // the last reading from each finger group [current code, last code, buttons pressed, max pressed]
int right[]   = {0, 0, 0, 0};
boolean repeatKeyMode = false; 
boolean shiftPressed = false;                  // Control keys
boolean altPressed = false;
boolean controlPressed = false;
boolean funPressed = false;
boolean space_L_Pressed = false;
boolean space_R_Pressed = false;
boolean space_L_Consumed = false;
boolean space_R_Consumed = false;
boolean pedal_L_Pressed = false;
boolean pedal_R_Pressed = false;
boolean winRequested = false;
boolean capitalRequested = false;

//
// Arrays to control arrow buttons
//
char arrow_buttons[] = {_CH0, _CH0, _CH0, _CH0};
const char arrowMap_0[] = {_LEFT,_UP,_DOWN,_RIGHT};
const char arrowMap_1[] = {_HOM,_PUP,_PDN,_END};
//long lastLPedalPress = 0;
//long lastRPedalPress = 0;

//
// Button assignments
//
const char funKeys[] = {_F01,_F02,_F03,_F04,_F05,_F06,_F07,_F08,_F09,_F10,_F11,_F12};
const char comKeys[] = {_HOM,_END,_PUP,_PDN,_INS,_RTN,_PRN};
const char ctrlKeys0[] = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'};
const char ctrlKeys1[] = {'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z'};

const char leftMapping[] =  {_CH0, 'e', 'a', 't', 'l', 'p', 's', 'c',_DEL, 'w',_CH0,_CH0,_TAB,_ESC, 'C', 'X',
                             _CH0, 'g', 'q', 'd', 'u', 'b', 'z', '*',_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_HOM,_CH0,
                             _CH0, '1', '2', '3', '4', '5', '6', '7', 'e',_CH0,_CH0,_CH0, '8', '9', '0', '.',
                             _CH0, '6', '7', '8', '9', '-', '0', '0', 'e',_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,
                             _CH0,_F05,_F04,_F06,_F03,_CH0,_F02,_PRN,_INS,_CH0,_CH0,_CH0,_F01,_CH0,_CPS,_CPS,
                             _CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0};
const char rightMapping[] = {_CH0, 'i', 'o', 'n', 'r', 'y', 'h', 'k',_BSP, 'q',_CH0,_CH0,_RTN,_CH0, 'V', 'Z',
                              ' ', '4', '5', '0', '6', 'z', 'x', 'x', ';',_CH0,_CH0,_CH0, ';',_CH0,_END,_CPS,
                             _CH0, ',', '.', '!', '?',_CH0, ':',_CH0, ';',_CH0,_CH0,_CH0, ';',_CH0,_CH0,_CH0,
                             _CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,
                             _CH0,_F07,_F08,_F10,_F09,_CH0,_F11,_F12,_CH0,_CH0,_CH0,_CH0,_INS,_CH0,_F05,_F06,
                             _CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0,_CH0};

const int long_dash[] = {0,1,5,1};
const int degree[] = {0,1,7,6};
const char keypad_buttons[] = {(char)234,(char)225,(char)226,(char)227,(char)228,(char)229,(char)230,(char)231,(char)232,(char)233};

//
// Setup code here, to run once:
//
void setup()
{
  #ifdef _DEBUG
  Serial.begin(9600);
  #endif

  pinMode(ledPin1, OUTPUT);
  pinMode(ledPin2, OUTPUT);
  digitalWrite(ledPin1, HIGH);
  digitalWrite(ledPin2, HIGH);

  Calibrate_Pots();

  for ( int i = 0; i < 4; i++)
  {
    pinMode(readPins[i], INPUT_PULLUP);
  }
  for ( int i = 0; i < 6; i++)
  {
    pinMode(writePins[i], OUTPUT);
    digitalWrite(writePins[i], HIGH);
  }

  ReadTopButtons();
  keyboardActive = left[2] + right[2] == 0;

  if ( keyboardActive)
  {
    Mouse.begin();
    Keyboard.begin();
    delay( 500);
    digitalWrite(ledPin1, LOW);
    digitalWrite(ledPin2, LOW);
    Keyboard.releaseAll();
  }
}

//
// runs continuously
//
void loop()
{
  int lCombo;
  int rCombo;
  boolean ProcessFurther;
  int mouseCounter = 0;
  if ( keyboardActive)
  {
    ReadBottomButtons();
    ReadArrowButtons();
    ReadTopButtons();
    if( mouseCounter == 0) Read_Mouse_Pots();
    if( (left[2] == 0) && (right[2] == 0) && ((left[3] > 0) || (right[3] > 0)))
    {
      #ifdef _DEBUG_KEYBOARD
      Serial.print( String( repeatKeyMode) + "_" + String( left[1]) + "_" + String( right[1]) + "\n");
      #endif
      if( (!repeatKeyMode) && DecodeTwoHanded())
      {
        lCombo = left[1];
        if( right[1] == 16) lCombo += 32;
        if( funPressed) lCombo += 64;
        rCombo = right[1];
        if( left[1] == 16) rCombo += 32;
        if( funPressed) rCombo += 64;
        DecodeButtons( lCombo, leftMapping);
        DecodeButtons( rCombo, rightMapping);
      }
      for( int i=0; i<4; i++)
      {
        left[i] = 0;
        right[i] = 0;
      }
      repeatKeyMode = false;
    }
    if( ((left[1] == 16) || (left[2] >= 2)) && (right[2] == 0) && (right[3] > 0))
    {
      rCombo = right[1];
      if( left[1] == 16) rCombo += 32;
      if( DecodeTwoHanded()) DecodeButtons( rCombo, rightMapping);
      for( int i=0; i<4; i++) right[i] = 0;
      repeatKeyMode = true;
    }
    if( ((right[1] == 16) || (right[2] >= 2)) && (left[2] == 0) && (left[3] > 0))
    {
      lCombo = left[1];
      if( right[1] == 16) lCombo += 32;
      if( DecodeTwoHanded()) DecodeButtons( lCombo, leftMapping);
      for( int i=0; i<4; i++) left[i] = 0;
      repeatKeyMode = true;
    }
    digitalWrite(ledPin1, capitalRequested || shiftPressed || funPressed);
    digitalWrite(ledPin2, altPressed || controlPressed || winRequested);
  }
  #ifdef _DEBUG
  delay(500);
  #else
  // the mouse reading is decimated
  mouseCounter++;
  if( mouseCounter >= _MOUSE_DECIMATOR) mouseCounter = 0;
  delay(5);
  #endif
  return;
}

//
// Calibrates middle points in all 4 pots, only needs to run once
//
void Calibrate_Pots()
{
  int pin, data;
  for( int i=0; i<4; i++)
  {
    pin = pot_sensorPins[i];
    data = analogRead(pin);
    delay(10);
    data += analogRead(pin);
    delay(10);
    data += analogRead(pin);
    delay(10);
    data += analogRead(pin);
    delay(10);
    data += analogRead(pin);
    pot_Middle[i] = data/5;
    pot_Position[i] = (signed char)pot_Middle[i];
  }
  return;
}

//
// Reads the pots and sends mouse commands
// Reads 3 mouse buttons using read lines 0,1,2 and write line 3
//
void Read_Mouse_Pots()
{
  int pin, data;
  bool button_L, button_R, button_M;

  //
  // Read mouse pots
  //
  for( int i=0; i<3; i++)
  {
    pin = pot_sensorPins[i];
    data = analogRead(pin);
    delay(3);
    data += analogRead(pin);
    data /= 2;
    if( data < pot_Middle[i])
      data = map(data, pot_Low[i], pot_Middle[i], pot_Calibrated_Low[i], 0);
    else
      data = map(data, pot_Middle[i], pot_High[i],  0, pot_Calibrated_High[i]);
    if( data < -10) pot_Position[i] = (signed char)(-data*data/30);
    if( -10 <= data && data < -7) pot_Position[i] = (signed char)(-2);
    if( -7 <= data && data < -4) pot_Position[i] = (signed char)(-1);
    if( -4 <= data && data <= 4) pot_Position[i] = (signed char)0;
    if( 4 < data && data <= 7) pot_Position[i] = (signed char)1;
    if( 7 < data && data <= 10) pot_Position[i] = (signed char)2;
    if( 10 < data) pot_Position[i] = (signed char)(data*data/30);
  }

  //
  // Read mouse buttons
  //
  digitalWrite( writePins[1], LOW);
  button_L = ReadPinLogical( 0);
  button_R = ReadPinLogical( 1);
  button_M = ReadPinLogical( 2);
  digitalWrite( writePins[1], HIGH);

  if( button_L && !Mouse.isPressed(MOUSE_LEFT)) Mouse.press(MOUSE_LEFT);
  if( !button_L && Mouse.isPressed(MOUSE_LEFT)) Mouse.release(MOUSE_LEFT);
  if( button_R && !Mouse.isPressed(MOUSE_RIGHT)) Mouse.press(MOUSE_RIGHT);
  if( !button_R && Mouse.isPressed(MOUSE_RIGHT)) Mouse.release(MOUSE_RIGHT);
  if( button_M && !Mouse.isPressed(MOUSE_MIDDLE)) Mouse.press(MOUSE_MIDDLE);
  if( !button_M && Mouse.isPressed(MOUSE_MIDDLE)) Mouse.release(MOUSE_MIDDLE);

  #ifndef _DISABLE_MOUSE
  if( pot_Position[0] != 0 || pot_Position[1] != 0 || pot_Position[2] != 0)
    Mouse.move(pot_Position[0], pot_Position[1], pot_Position[2]>>1);
  #endif

  #ifdef _DEBUG_MOUSE
  Serial.println( "Mouse pointer: " + String(pot_Position[0]) + " "
    + String(pot_Position[1]) + " " + String(pot_Position[2]) + " " + String(pot_Position[3]));
  #endif
  return;
}

//
// Reads 4 arrow buttons using read lines 0,1,2,3 and write line 1
//
void ReadArrowButtons()
{
  bool button;
  digitalWrite( writePins[3], LOW);
  for( int i=0; i<4; i++)
  {
    button = ReadPinLogical( i);
    if( !button && arrow_buttons[i] != 0)
    {
      Keyboard.release( arrow_buttons[i]);
      arrow_buttons[i] = _CH0;
    }
    if( !space_L_Pressed && button && arrow_buttons[i] != arrowMap_0[i])
    {
      space_L_Consumed = false;
      if( arrow_buttons[i] != 0) Keyboard.release( arrow_buttons[i]);
      Keyboard.press( arrowMap_0[i]);
      arrow_buttons[i] = arrowMap_0[i];
    }
    if( space_L_Pressed && button && arrow_buttons[i] != arrowMap_1[i])
    {
      space_L_Consumed = true;
      if( arrow_buttons[i] != 0) Keyboard.release( arrow_buttons[i]);
      Keyboard.press( arrowMap_1[i]);
      arrow_buttons[i] = arrowMap_1[i];
    }
  }
  digitalWrite( writePins[3], HIGH);
  return;
}

//
// Reads top 8 buttons using control lines 0 and 1 and data lines 0,1,2,3
// Adds the Space buttons using data lines 2, 3, while control sinks to 1
//
void ReadTopButtons()
{
    left[0] = 0;
    right[0] = 0;
    left[2] = 0;
    right[2] = 0;
    int mult = 1;
    int b;

    // Left half
    digitalWrite(writePins[0], LOW);
    for( int i=0; i<4; i++)
    {
      b = ReadPin( i);
      left[0] += mult * b;
      left[2] += b;
      mult *= 2;
    }
    digitalWrite(writePins[0], HIGH);
    if( space_L_Pressed)
    {
      left[0] += 16;
      left[2] += 1;
    }
    //if( funPressed) left[0] += 32;
    if ( left[2] >= left[3])
    {
      left[1] = left[0];
      left[3] = left[2];    
    }

    // Right half
    mult = 1;
    digitalWrite(writePins[2], LOW);
    for( int i=0; i<4; i++)
    {
      b = ReadPin( i);
      right[0] += mult * b;
      right[2] += b;
      mult *= 2;
    }
    digitalWrite(writePins[2], HIGH);
    if( space_R_Pressed)
    {
      right[0] += 16;
      right[2] += 1;
    }
    //if( funPressed) right[0] += 32;
    if ( right[2] >= right[3])
    {
      right[1] = right[0];
      right[3] = right[2];    
    }
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Top buttons: " + String( left[0]) + ", " + String( right[0]) + "\n");
    #endif
    return;
}

//
// Reads bottom 6 buttons and pedals using control lines 4,5 and data lines 1,2,3,4
// Sends shift, alt, control, sets function and space
//
void ReadBottomButtons()
{
  boolean button2;
  boolean button3;

  // Left half
  digitalWrite(writePins[4], LOW);
  space_L_Pressed = ReadPinLogical( 0);
  button2 = ReadPinLogical( 1);
  button3 = ReadPinLogical( 2);
  bool pedal = ReadPinLogical( 3);
  digitalWrite(writePins[4], HIGH);
  if ( button2 && (!shiftPressed))
  {
    Keyboard.press(_SHIFT);
    shiftPressed = true;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Shift pressed\n");
    #endif
  }
  if ( (!button2) && shiftPressed)
  {
    Keyboard.release(_SHIFT);
    shiftPressed = false;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Shift released\n");
    #endif
  }
  if ( button3 && (!controlPressed))
  {
    Keyboard.press(_CTRL);
    controlPressed = true;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Control pressed\n");
    #endif
  }
  if ( (!button3) && controlPressed)
  {
    Keyboard.release(_CTRL);
    controlPressed = false;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Control released\n");
    #endif
  }
  if( !pedal && pedal_L_Pressed)
  {
    pedal_L_Pressed = false;
  }
  if( pedal && !pedal_L_Pressed)
  {
    capitalRequested = !capitalRequested;
    pedal_L_Pressed = true;
    digitalWrite(ledPin1, capitalRequested || shiftPressed || funPressed);
  }

  // Right half
  digitalWrite(writePins[5], LOW);
  space_R_Pressed = ReadPinLogical( 0);
  button2 = ReadPinLogical( 1);
  button3 = ReadPinLogical( 2);
  pedal = ReadPinLogical( 3);
  digitalWrite(writePins[5], HIGH);
  if ( button2 && (!funPressed))
  {
    funPressed = true;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Fun pressed\n");
    #endif
  }
  if ( (!button2) && funPressed)
  {
    funPressed = false;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Fun released\n");
    #endif
  }
  if ( button3 && (!altPressed))
  {
    Keyboard.press(_ALT);
    altPressed = true;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Alt pressed\n");
    #endif
  }
  if ( (!button3) && altPressed)
  {
    Keyboard.release(_ALT);
    altPressed = false;
    #ifdef _DEBUG_KEYBOARD
    Serial.print( "Alt released\n");
    #endif
  }
  if( !pedal && pedal_R_Pressed)
  {
    pedal_R_Pressed = false;
  }
  if( pedal && !pedal_R_Pressed)
  {
    if( !controlPressed) Keyboard.press(_CTRL);
    if( !shiftPressed) Keyboard.press(_SHIFT);
    if( !shiftPressed) Keyboard.release(_SHIFT);
    if( !controlPressed) Keyboard.release(_CTRL);
    pedal_R_Pressed = true;
  }
  return;
}

//
// Decodes the pressed button for two-handed chords
// Returns true if further processing is not requred
//
boolean DecodeBaseChord( int l, int r, char unshifted)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  if( space_L_Consumed)
  {
     space_L_Consumed = false;
     return true;
  }
  if( winRequested) Keyboard.press( _GUI);
  if( capitalRequested) Keyboard.press( _SHIFT);
  Keyboard.write( unshifted);
  if( capitalRequested) Keyboard.release( _SHIFT);
  if( winRequested) Keyboard.release( _GUI);
  ReleaseAllControls();
  return true;
}

//
// Decodes the pressed button for two-handed chords
// Returns true if further processing is not requred
// Makes substitutes '@' - '#' and '^' - '&' for correct show of Ъ and Ё
//
boolean DecodeRussianChord( int l, int r, char unshifted, char shifted)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  if( shiftPressed || capitalRequested) Keyboard.write( shifted);
  else Keyboard.write( unshifted);
  ReleaseAllControls();
  return true;
}

//
// Decodes the pressed button for two-handed chords
// Returns true if further processing is not requred
// Converts chord into a control-sequience
//
boolean DecodeControlChord( int l, int r, char c)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  ReleaseAllControls();
  SendControl(c);
  ReleaseAllControls();
  return true;
}

//
// Decodes the pressed button for two-handed chords
// Returns true if further processing is not requred
// Outputs a string-sequience
//
boolean DecodeStringChord( int l, int r, String s)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  int len = s.length();
  if( len <= 0) return false;
  if( capitalRequested) Keyboard.press( _SHIFT);
  Keyboard.write( s[0]);
  if( capitalRequested) Keyboard.release( _SHIFT);
  ReleaseAllControls();
  for( int i=1; i<len; i++) Keyboard.write( s[i]);  
  return true;
}

//
// Decodes the triple-dot into a single char
// Returns true if further processing is not requred
//
boolean DecodeTripleDot( int l, int r)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  Keyboard.print( " ... ");
  Keyboard.write( _LEFT);  
  Keyboard.write( _LEFT);  
  Keyboard.write( _BSP);  
  Keyboard.write( _RIGHT);  
  Keyboard.write( _RIGHT);  
  return true;
}

//
// Decodes the pressed button for two-handed chords
// Returns true if further processing is not requred
// Outputs a couple of brackets and positions cursor between them
//
boolean DecodeArrowChord( int l, int r, String s, char Mv, int rpt)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  int len = s.length();
  if( len <= 0) return false;
  Keyboard.print( s);
  for( int i=0; i<rpt; i++) Keyboard.write( Mv);  
  return true;
}

//
// Decodes the pressed button for chords followed by return
//
boolean DecodeReturnStringChord( int l, int r, String s)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  int len = s.length();
  if( len <= 0) return false;
  if( capitalRequested) Keyboard.press( _SHIFT);
  Keyboard.write( s[0]);
  if( capitalRequested) Keyboard.release( _SHIFT);
  ReleaseAllControls();
  for( int i=1; i<len; i++) Keyboard.write( s[i]);  
  Keyboard.write( _RTN);
  return true;
}

//
// Decodes the Win key request
// Returns true if further processing is not requred
//
boolean DecodeWinChord( int l, int r)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  if( winRequested)
  {
    Keyboard.press( _GUI);
    Keyboard.release( _GUI);
    winRequested = false;
  }
  else
  {
    winRequested = true;
  }
  return true;
}

//
// Decodes the scan codes for sending special characters
// The length must be 4 characters
//
boolean DecodeKeypad( int l, int r, const int kp[])
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  Keyboard.press( _ALT);
  for( int i=0; i<4; i++)
  {
    Keyboard.write( keypad_buttons[kp[i]]);
  }
  Keyboard.release( _ALT);
  return true;
}

//
// Decodes double space for sending a capital letter after space or return
//
boolean DecodeCapital( int l, int r, int ch)
{
  if( left[1] != l) return false;
  if( right[1] != r) return false;
  Keyboard.write( ch);
  capitalRequested = true;
  return true;
}

//
// Decodes all two-handed chords
// Returns true if further processing is requred
//
boolean DecodeTwoHanded()
{
  // switch is used to speed up processing
  switch( left[1])
  {
    case 0:
      if( DecodeBaseChord(  0,16, ' ')) return false;
      if( DecodeReturnStringChord( 0, 28, ";")) return false;
      return true;
    case 1:
      if( DecodeBaseChord( 1, 1, 'f')) return false;
      if( DecodeBaseChord( 1, 2, 'm')) return false;
      if( DecodeBaseChord( 1, 4, 'd')) return false;
      if( DecodeArrowChord( 1, 5, "\"\"", _LEFT, 1)) return false;
      if( DecodeStringChord( 1, 15, "ed")) return false;
      if( DecodeBaseChord( 1, 3, 'g')) return false;
      if( DecodeBaseChord( 1, 6, '-')) return false;
      if( DecodeBaseChord( 1, 7, '`')) return false;
      if( DecodeRussianChord( 1, 14, '>', '>')) return false;
      if( DecodeBaseChord( 1, 19, '7')) return false;
      return true;
    case 2:
      if( DecodeBaseChord( 2, 1, 'j')) return false;
      if( DecodeBaseChord( 2, 2, 'b')) return false;
      if( DecodeBaseChord( 2, 3, 'q')) return false;
      if( DecodeArrowChord( 2, 5, "\'\'", _LEFT, 1)) return false;
      if( DecodeBaseChord( 2, 6, '/')) return false;
      if( DecodeBaseChord( 2, 7, '\\')) return false;
      if( DecodeRussianChord( 2, 14, '<', '<')) return false;
      if( DecodeStringChord( 2, 15, "ad")) return false;
      if( DecodeBaseChord( 2, 19, '8')) return false;
      return true;
    case 3:
      if( DecodeBaseChord( 3, 1, 'v')) return false;
      if( DecodeBaseChord( 3, 2, 'w')) return false;
      if( DecodeRussianChord( 3, 3, '"', '"')) return false;
      if( DecodeBaseChord( 3, 4, 'z')) return false;
      if( DecodeStringChord( 3, 6, "th")) return false;
      if( DecodeArrowChord( 3, 14, "if()", _LEFT, 1)) return false;
      if( DecodeStringChord( 3, 15, "un")) return false;
      return true;
    case 4:
      if( DecodeBaseChord( 4, 1, 'u')) return false;
      if( DecodeBaseChord( 4, 4, 'x')) return false;
      if( DecodeStringChord( 4, 5, "ly")) return false;
      if( DecodeBaseChord( 4, 6, '*')) return false;
      if( DecodeArrowChord( 4, 14, "for( int i=0; i<; i++)", _LEFT, 6)) return false;
      if( DecodeStringChord( 4, 15, "al")) return false;
      if( DecodeBaseChord( 4, 19, '9')) return false;
      return true;
    case 5:
      if( DecodeBaseChord( 5, 1, '[')) return false;
      if( DecodeBaseChord( 5, 2, ']')) return false;
      if( DecodeRussianChord( 5, 4, '^', '&')) return false;
      if( DecodeStringChord( 5, 5, ";\n")) return false;
      if( DecodeStringChord( 5, 6, "ph")) return false;
      if( DecodeArrowChord( 5, 14, "foreach(  in)", _LEFT, 4)) return false;
      return true;
    case 6:
      if( DecodeRussianChord( 6, 1, '(', '<')) return false;
      if( DecodeRussianChord( 6, 2, ')', '>')) return false;
      if( DecodeStringChord( 6, 3, "sh")) return false;
      if( DecodeRussianChord( 6, 6, '\'', '\'')) return false;
      if( DecodeArrowChord( 6, 14, "for i in range( len()):", _LEFT, 3)) return false;
      if( DecodeStringChord( 6, 15, "and")) return false;
      return true;
    case 7:
      if( DecodeBaseChord( 7, 1, '=')) return false;
      if( DecodeRussianChord( 7, 2, '@', '#')) return false;
      if( DecodeStringChord( 7, 3, "qu")) return false;
      if( DecodeRussianChord( 7, 4, '^', '&')) return false;
      if( DecodeStringChord( 7, 6, "ch")) return false;
      if( DecodeStringChord( 7, 7, "ck")) return false;
      if( DecodeArrowChord( 7, 14, "print( )", _LEFT, 1)) return false;
      return true;
    case 11:
      if( DecodeStringChord( 11, 1, _PASSWORD_1)) return false;
      if( DecodeStringChord( 11, 2, _PASSWORD_2)) return false;
      if( DecodeStringChord( 11, 3, _PASSWORD_3)) return false;
      if( DecodeStringChord( 11, 4, _PASSWORD_4)) return false;
      if( DecodeStringChord( 11, 5, _PASSWORD_5)) return false;
      if( DecodeStringChord( 11, 6, _PASSWORD_6)) return false;
      if( DecodeStringChord( 11, 7, _PASSWORD_7)) return false;
      if( DecodeStringChord( 11, 8, _PASSWORD_8)) return false;
      if( DecodeStringChord( 11, 9, _PASSWORD_9)) return false;
      if( DecodeStringChord( 11, 10, _PASSWORD_10)) return false;
      if( DecodeStringChord( 11, 11, _PASSWORD_11)) return false;
      if( DecodeStringChord( 11, 12, _PASSWORD_12)) return false;
      if( DecodeStringChord( 11, 13, _PASSWORD_13)) return false;
      if( DecodeStringChord( 11, 14, _PASSWORD_14)) return false;
      if( DecodeStringChord( 11, 15, _PASSWORD_15)) return false;
      return true;
    case 12:
      if( DecodeStringChord( 12, 1, "+")) return false;
      if( DecodeStringChord( 12, 2, "-")) return false;
      if( DecodeStringChord( 12, 3, "&")) return false;
      if( DecodeStringChord( 12, 4, "*")) return false;
      if( DecodeStringChord( 12, 6, "|")) return false;
      if( DecodeStringChord( 12, 7, "$")) return false;
      if( DecodeStringChord( 12, 8, "/")) return false;
      if( DecodeBaseChord( 12, 12, _ESC)) return false;
      return true;
    case 14:
      if( DecodeArrowChord( 14, 1, "()", _LEFT, 1)) return false;
      if( DecodeArrowChord( 14, 2, "[]", _LEFT, 1)) return false;
      if( DecodeArrowChord( 14, 3, "<>", _LEFT, 1)) return false;
      if( DecodeArrowChord( 14, 4, "{}", _LEFT, 1)) return false;
      if( DecodeArrowChord( 14, 6, "\n{\n\n}", _UP, 1)) return false;
      if( DecodeStringChord( 14, 8, "ou")) return false;
      if( DecodeStringChord( 14, 12, ":\n}")) return false;
      if( DecodeStringChord( 14, 14, "the")) return false;
      return true;
    case 15:
      if( DecodeStringChord( 15, 1, "in")) return false;
      if( DecodeStringChord( 15, 2, "on")) return false;
      if( DecodeStringChord( 15, 3, "ing")) return false;
      if( DecodeStringChord( 15, 4, "er")) return false;
      if( DecodeStringChord( 15, 6, "or")) return false;
      if( DecodeControlChord( 15, 15, 'Z')) return false;
      return true;
    case 16:
      if( DecodeBaseChord( 16, 0, _BSP)) return false;
      //if( DecodeStringChord( 16, 7, "... ")) return false;
      if( DecodeTripleDot( 16, 7)) return false;
      if( DecodeReturnStringChord( 16, 8, ";")) return false;
      if( DecodeCapital( 16, 12, _RTN)) return false;
      if( DecodeKeypad( 16, 14, long_dash)) return false;
      if( DecodeKeypad( 16, 15, degree)) return false;
      if( DecodeCapital( 16, 16, ' ')) return false;
      if( DecodeCapital( 16, 24, _RTN)) return false;
      return true;
    case 17:
      if( DecodeStringChord( 17, 17, "#")) return false;
      return true;
    case 18:
      if( DecodeStringChord( 18, 2, "who")) return false;
      if( DecodeStringChord( 18, 6, "wh")) return false;
      return true;
    case 19:
      if( DecodeStringChord( 19, 1, "t=")) return false;
      return true;
    case 30:
      if( DecodeStringChord( 30, 16, "00")) return false;
      return true;
    case 31:
      if( DecodeWinChord( 31, 0)) return false;
      if( DecodeStringChord( 31, 16, "000")) return false;
      return true;
  }//switch
  return true;
}

//
// Decodes the pressed button
//
void DecodeButtons( int n, const char Mapping[])
{
  if( n <= 0) return;
  if( n > 79) n = 79;
  char c = Mapping[n];
  c = SendControl( c);
  if( c == ' ') return;
  if( winRequested) Keyboard.press( _GUI);
  if( capitalRequested) Keyboard.press( _SHIFT);
  Keyboard.write( c);
  if( capitalRequested) Keyboard.release( _SHIFT);
  if( winRequested) Keyboard.release( _GUI);
  ReleaseAllControls();
}

//
// Converts button into a Control-Sequence
// Returns space if sent 
//
char SendControl( char c)
{
  for( int i=0; i<26; i++)
  {
    if( c != ctrlKeys0[i]) continue;
    if( controlPressed)
    {
      Keyboard.write( ctrlKeys1[i]);
      return ' ';
    }
    Keyboard.press(_CTRL);
    Keyboard.write( ctrlKeys1[i]);
    Keyboard.release(_CTRL);
    return ' '; 
  }
  return c;
}

//
// Returns true if the character is a function key
//
boolean IsFunKey( char c)
{
  for( int i=0; i<12; i++)
  {
    if( c == funKeys[i]) return true;
  }
  for( int i=0; i<7; i++)
  {
    if( c == comKeys[i]) return true;
  }
  return false;
}

//
// Releases Control, Alt, Shift, Win if pressed 
//
void ReleaseAllControls()
{
    if( controlPressed)
    {
      controlPressed = false;
      Keyboard.release( _CTRL);
    }
    if( altPressed)
    {
      altPressed = false;
      Keyboard.release( _ALT);
    }
    if( shiftPressed)
    {
      shiftPressed = false;
      Keyboard.release( _SHIFT);
    }
    funPressed = false;
    winRequested = false;
    capitalRequested = false;
}

//
// Reads a pin, returns the reading
//
int ReadPin( int i)
{
  if ( digitalRead(readPins[i]) == LOW) return 1;
  return 0;
}

//
// Reads a pin, returns the reading
//
int ReadPinLogical( int i)
{
  return digitalRead(readPins[i]) == LOW;
}

